' File: ReflectElemsPropsMtlCsys.BAS
' Version: Version 1
' Date: June 28, 2012
' Author: Michael Saeger, Kurt Knutson, ATA Engineering
' www.ata-plmsoftware.com
' Description: This macro copies and reflects elements across a symmetry plane.
'	Offsets, orientations, properties, and coordinate systems are reflected along
'	with the elements so that reflected elements have the same relative alignment.
'
'Usage Notes
'====================================
' Coordinate systems cannot be reflected perfectly because FEMAP does not support left-handed 
' coordinate systems. For this reason, X- and Y-axes are reflected and the Z-axis is made 
' perpendicular to both in a right-handed orientation.
'
'Copyright (c) <2012> <ATA Engineering, Inc.>
'
'Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
'associated documentation files (the "Software") to use, copy, modify, merge, and/or distribute
'royalty-free copies of the Software, and to permit persons to whom the Software is furnished
'to do so, subject to the following conditions:
'The above copyright notice and this permission notice shall be included in all copies or
'substantial portions of the Software.
'THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
'INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
'FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
'WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
'OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.       
' 
                                'Set a constant to control whether debug information will be output to the text window
Const DebugMode As Boolean = False
                                'Set a constant for the largest possible ID number
Const MaxAllowableID As Long = 99999999
                                'Set a length tolerance value for unit vectors to avoid possible crash due to roundoff error
Const unitlength_tolerance As Double = 1E-9
                                'Set a constant of type double with a value of unity
Const unity As Double = 1.0
								'Define a string indicating the version of the translator
Const API_version As String = "1"

Sub Main
                                'Connect to the FEMAP model
Dim App As femap.model
Set App = GetObject(,"femap.model")
								'Indicate which version and contact information for support
App.feAppMessage(FCM_NORMAL,"*** Write Out Sets, Version " & API_version & " ***")
App.feAppMessage(FCM_NORMAL,"Created by ATA Engineering, http://www.ata-plmsoftware.com")
                                'Create a Set object to store the elements that will be reflected
Dim ElemsToReflect As femap.Set
Set ElemsToReflect = App.feSet
                                'Create a Set object to store the elements referenced by selected weld elements
Dim ElemsOnWeldsToReflect As femap.Set
Set ElemsOnWeldsToReflect = App.feSet
                                'Create a Set object to store the elements referenced by selected weld elements
Dim WeldElementsToRemove As femap.Set
Set WeldElementsToRemove = App.feSet
                                'Create a node indexing variable for weld elements
Dim WeldElementNodeIndex As Long
                                'Create a Set object to store the nodes that will be reflected
Dim NodesToReflect As femap.Set
Set NodesToReflect = App.feSet
                                'Create a Set object to store the properties that will be reflected
Dim PropertiesToReflect As femap.Set
Set PropertiesToReflect = App.feSet
                                'Create a boolean variable to store whether or not to use an element type filter
Dim UseElemTypeFilter As Boolean
                                'Create a Set object to store the element types filter that will restrict reflected property creation
Dim ElemTypeFilter As femap.Set
Set ElemTypeFilter = App.feSet
                                'Create a Set object to store the elements whose types will be a filter that will restrict reflected property creation
Dim ElemTypeFilterElements As femap.Set
Set ElemTypeFilterElements = App.feSet
                                'Create a boolean variable to store whether or not to use a material type filter
Dim UseMaterialTypeFilter As Boolean
                                 'Create a Set object to store the material types filter that will restrict reflected property creation
Dim MaterialTypeFilter As femap.Set
Set MaterialTypeFilter = App.feSet
                                'Create a Set object to store the elements whose material types will be a filter that will restrict reflected property creation
Dim MaterialTypeFilterElements As femap.Set
Set MaterialTypeFilterElements = App.feSet
                                'Create a boolean variable to store whether or not an element with no material will pass the material type filter
Dim IncludeNoMaterialInMaterialTypeFilter As Boolean
                                'Create a boolean variable to store whether or not an element passes the material type filter
Dim MaterialTypeFilterPassed As Boolean
                                'Create a Set object to store the coordinate systems that will be reflected
Dim CSysToReflect As femap.Set
Set CSysToReflect = App.feSet
                                'The ID of zero should not be added to the Set object storing the coordinate systems that will be reflected so use a boolean instead for this ID only
Dim ReflectCS0 As Boolean
                                'Create a Sort object to store a table of original node IDs and reflected node IDs
Dim ReflectionNodeLookup As femap.SortSet
Set ReflectionNodeLookup = App.feSort
                                'Create a Sort object to store a table of original property IDs and reflected property IDs
Dim ReflectionPropertyLookup As femap.SortSet
Set ReflectionPropertyLookup = App.feSort
                                'Create a Sort object to store a table of original CSys IDs and reflected CSys IDs
Dim ReflectionCSysLookup As femap.SortSet
Set ReflectionCSysLookup = App.feSort
                                'The ID of zero should not be added to the Sort object storing the coordinate system ID lookups so use an integer instead for this ID only
Dim ReflectedCS0ID As Long
                                'Create an Elem object to access element data
Dim ThisElem As femap.Elem
Set ThisElem = App.feElem
                                'Create a Node object to access node data
Dim ThisNode As femap.Node
Set ThisNode = App.feNode
                                'Create a Property object to access property data
Dim ThisProperty As femap.Prop
Set ThisProperty = App.feProp
                                'Create a Matl object to access material data
Dim ThisMaterial As femap.Matl
Set ThisMaterial = App.feMatl
                                'Create a CSys object to access coordinate system data
Dim ThisCsys As femap.CSys
Set ThisCsys = App.feCSys
                                'Create an Elem object to check if a particular element ID is available
Dim TestElemID As femap.Elem
Set TestElemID = App.feElem
                                'Create a Node object to check if a particular node ID is available
Dim TestNodeID As femap.Node
Set TestNodeID = App.feNode
                                'Create a Property object to check if a particular property ID is available
Dim TestPropertyID As femap.Prop
Set TestPropertyID = App.feProp
                                'Create a CSys object to check if a particular coordinate system ID is available
Dim TestCsysID As femap.CSys
Set TestCsysID = App.feCSys
                                'Create a Set object to store the nodes on a single element that will be reflected
Dim NodesToReflectForThisElem As femap.Set
Set NodesToReflectForThisElem = App.feSet
                                'Create an integer variable to store the ID offset that will be assigned to new nodes
Dim NewNodeOffsetID As Long
                                'Create an integer variable to store the ID offset that will be assigned to new elements
Dim NewElemOffsetID As Long
                                'Create an integer variable to store the ID offset that will be assigned to new properties
Dim NewPropOffsetID As Long
                                'Create an integer variable to store the ID offset that will be assigned to new coordinate systems
Dim NewCSysOffsetID As Long
                                'Create an integer variable to store the ID value that will be assigned to new nodes
Dim NewNodeID As Long
                                'Create an integer variable to store the ID value that will be assigned to new elements
Dim NewElemID As Long
                                'Create an integer variable to store the ID value that will be assigned to new properties
Dim NewPropID As Long
                                'Create an integer variable to store the ID value that will be assigned to new coordinate systems
Dim NewCSysID As Long
                                'Create a three-element array of double precision floating points to store the global origin
Dim GlobalOrigin(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store a point on the mirror plane
Dim SymPlanePoint(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store the normal vector of the mirror plane
Dim SymPlaneNormal(0 To 2) As Double
                                'Create a variant to store a point on the mirror plane
Dim vSymPlanePoint As Variant
                                'Create a variant to store the normal vector of the mirror plane
Dim vSymPlaneNormal As Variant
                                'Create a variant to store a dummy vector argument
Dim vDummyVector As Variant
                                'Create a boolean variable to store whether or not to reflect the output and definition coordinate systems of the nodes
Dim ReflectNodeCSys As Boolean
                                'Create a three-element array of double precision floating points to store the coordinates of a point to be reflected
Dim PointToReflect(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store the coordinates of a point after it has been reflected
Dim ReflectedPoint(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store the x-axis to be reflected
Dim XaxisToReflect(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store the y-axis to be reflected
Dim YaxisToReflect(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store the x-axis after it has been reflected
Dim ReflectedXaxis(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store the y-axis after it has been reflected
Dim ReflectedYaxis(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store the z-axis after it has been reflected
Dim ReflectedZaxis(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store the x- then y- then z-axis rotations of a coordinate system
Dim CsysXYZrotations(0 To 2) As Double
                                'Create a set to store the node indeces for each element that will need to be reflected
Dim NodeIndeces As femap.Set
Set NodeIndeces = App.feSet
                                'Create an array to store the node indeces for each element that will need to be reflected with node order reversed
Dim NodeIndecesArray As Variant
                                'Create an integer to store the current location in an array
Dim ArrayPosition As Long
                                'Create an integer to store a dummy size argument
Dim DummySize As Long
                                'Create an array to use as a copy the original nodes on an element whose node order will be reversed
Dim OriginalNodeArray As Variant
                                'Create arrays to store node list information
Dim NumberOfNodesInNodeList As Long
Dim NodeListNodeArray As Variant
Dim NodeListFaceArray As Variant
Dim NodeListWeightArray As Variant
Dim NodeListDOFarray As Variant
                                'Create a three-element array of double precision floating points to store a general vector to be reflected
Dim VectorToReflect(0 To 2) As Double
                                'Create a three-element array of double precision floating points to store a general vector after it has been reflected
Dim ReflectedVector(0 To 2) As Double
                                'Create a boolean variable to indicate if there was a an error preventing an element from being reflected
Dim ReflectionError As Boolean
                                'Create a counter for warning messages
Dim WarningMsgCount As Long
                                'Initialize the counter for warning messages to zero
WarningMsgCount = 0
                                'Assign the global origin to (0, 0, 0)
GlobalOrigin(0) = 0.0
GlobalOrigin(1) = 0.0
GlobalOrigin(2) = 0.0
                                'Get the elements to reflect from the user
ElemsToReflect.Clear
                                'Ask the user which elements to reflect
ret = ElemsToReflect.Select(FT_ELEM,True,"Please select the elements you would like to reflect")
If (ret <> FE_CANCEL) Then
                                'Ask the user whether an element type filter should be used
  ret = App.feAppMessageBox(3,"All selected elements will be reflected, however you may wish to restrict the creation of reflected properties to certain element types. Use element type filter?" & vbCrLf & vbCrLf & "Yes - use an element type filter to restrict which elements will have reflected properties" & vbCrLf & "No - create reflected properties for all element types")
  UseElemTypeFilter = (ret = FE_OK)
  If (UseElemTypeFilter) Then
  	ElemTypeFilterElements.Clear
                                'If the user opts to use an element type filter then ask for elements that are of the types that will pass the filter
    ret = ElemTypeFilterElements.Select(FT_ELEM,True,"Please select a set of elements whose types will be a filter for reflected property creation")
    If (ret <> FE_CANCEL) Then
      ElemTypeFilter.Clear
                                'Add the selected element types to the filter
      ret = ElemTypeFilterElements.First
      While (ret <> FE_FAIL)
        ret = ThisElem.Get(ElemTypeFilterElements.CurrentID)
        If (ElemTypeFilter.IsAdded(ThisElem.type) = FE_FAIL) Then
          ret = ElemTypeFilter.Add(ThisElem.type)
        End If
        ret = ElemTypeFilterElements.Next
      Wend
    End If
  End If
End If
If (ret <> FE_CANCEL) Then
                                'Ask the user whether an element material type filter should be used
  ret = App.feAppMessageBox(3,"All selected elements will be reflected, however you may wish to restrict the creation of reflected properties to certain material types. Use material type filter?" & vbCrLf & vbCrLf & "Yes - use a material type filter to restrict which elements will have reflected properties" & vbCrLf & "No - create reflected properties for all material types")
  UseMaterialTypeFilter = (ret = FE_OK)
  If (UseMaterialTypeFilter) Then
  	MaterialTypeFilterElements.Clear
                                'If the user opts to use an element material type filter then ask for elements whose material types will pass the filter
    ret = MaterialTypeFilterElements.Select(FT_ELEM,True,"Please select a set of elements whose material types will be a filter for reflected property creation")
    If (ret <> FE_CANCEL) Then
      MaterialTypeFilter.Clear
                                'Add the selected element material types to the filter
      ret = MaterialTypeFilterElements.First
      While (ret <> FE_FAIL)
        ret = ThisElem.Get(MaterialTypeFilterElements.CurrentID)
        If ((ThisElem.propID > 0) And (ThisElem.propID <= MaxAllowableID)) Then
          If ((ThisProperty.matlID > 0) And (ThisProperty.matlID <= MaxAllowableID)) Then
            ret = ThisMaterial.Get(ThisProperty.matlID)
            If (MaterialTypeFilter.IsAdded(ThisMaterial.type) = FE_FAIL) Then
              ret = MaterialTypeFilter.Add(ThisMaterial.type)
            End If
          End If
        End If
        ret = MaterialTypeFilterElements.Next
      Wend
                                'Ask the user whether to include elements with no material in the filter
      ret = App.feAppMessageBox(3,"Would you like elements with no material to be included in the material type filter?" & vbCrLf & vbCrLf & "Yes - elements with no material will pass the material type filter" & vbCrLf & "No - elements with no material will not pass the material type filter" & vbCrLf & "Cancel - abort reflection")
      IncludeNoMaterialInMaterialTypeFilter = (ret = FE_OK)
    End If
  End If
End If
If (ret <> FE_CANCEL) Then
                                'Ask the user whether or not to reflect nodal coordinate systems
  ret = App.feAppMessageBox(3,"Would you like to create reflected coordinate systems for reflected nodes?" & vbCrLf & vbCrLf & "Yes - reflected nodes will reference reflected coordinate systems" & vbCrLf & "No - reflected nodes will reference the same coordinate systems as the original nodes" & vbCrLf & "Cancel - abort reflection")
  ReflectNodeCSys = (ret = FE_OK)
End If
If (ret <> FE_CANCEL) Then
                                'Ask the user for an ID offset to use for reflected nodes, or a flag (zero) to indicate next available ID
  NewNodeOffsetID = 0
  ret = App.feGetInt("Please enter offset for nodes (zero for next available)",0,MaxAllowableID-App.Info_MaxID(FT_NODE),NewNodeOffsetID)
End If
If (ret <> FE_CANCEL) Then
                                'Ask the user for an ID offset to use for reflected elements, or a flag (zero) to indicate next available ID
  NewElemOffsetID = NewNodeOffsetID
  ret = App.feGetInt("Please enter offset for elements (zero for next available)",0,MaxAllowableID-App.Info_MaxID(FT_ELEM),NewElemOffsetID)
End If
If (ret <> FE_CANCEL) Then
                                'Ask the user for an ID offset to use for reflected properties, or a flag (zero) to indicate next available ID
  NewPropOffsetID = NewElemOffsetID
  ret = App.feGetInt("Please enter offset for properties (zero for next available)",0,MaxAllowableID-App.Info_MaxID(FT_PROP),NewPropOffsetID)
End If
If (ret <> FE_CANCEL) Then
                                'Ask the user for an ID offset to use for reflected coordinate systems, or a flag (less than 3) to indicate next available ID
  NewCSysOffsetID = NewPropOffsetID
  ret = App.feGetInt("Please enter offset for coordinate systems (< 3 for next available)",0,MaxAllowableID-App.Info_MaxID(FT_CSYS),NewCSysOffsetID)
End If
If (ret <> FE_CANCEL) Then
  If (ElemsToReflect.Count = 0) Then
    App.feAppMessage(FCM_NORMAL,"No elements were selected.")
  Else
                                'Ask the user to select a reflection plane
    ret = App.fePlanePick("Please select a plane for reflection",vSymPlanePoint,vSymPlaneNormal,vDummyVector)
    If (ret <> FE_CANCEL) Then
      SymPlanePoint(0) = vSymPlanePoint(0)
      SymPlanePoint(1) = vSymPlanePoint(1)
      SymPlanePoint(2) = vSymPlanePoint(2)
      SymPlaneNormal(0) = vSymPlaneNormal(0)
      SymPlaneNormal(1) = vSymPlaneNormal(1)
      SymPlaneNormal(2) = vSymPlaneNormal(2)
                                'Look at all of the weld elements that were selected to see if they reference other elements
      ElemsOnWeldsToReflect.Clear
      WeldElementsToRemove.Clear
      ret = ElemsToReflect.First
      While(ret <> FE_FAIL)
        ret = ThisElem.Get(ElemsToReflect.CurrentID)
        If (ThisElem.topology = FTO_WELD2) Then
                                'Use the type of weld element to determine which elements and properties (if any) must be reflected along with the weld
          Select Case ThisElem.connectTYPE(0)
            Case 0,3,7 'Elem to Elem, Patch to Patch, Fastener to Elem
              ret = ElemsOnWeldsToReflect.Add(ThisElem.connectSEG(0))
              ret = ElemsOnWeldsToReflect.Add(ThisElem.connectSEG(1))
            Case 1 'Elem to Elem Vertex
              ret = ElemsOnWeldsToReflect.Add(ThisElem.connectSEG(0))
            Case 4,8 'Prop to Prop, Fastener to Prop
                                'Check the element type to see if it passes the filter for property reflection
              If (CheckFilter(ThisElem.type,UseElemTypeFilter,ElemTypeFilter)) Then
                ret = ThisProperty.Get(ThisElem.propID)
                                'Check the element material type to see if it passes the filter for property reflection
                If ((ThisProperty.matlID > 0) And (ThisProperty.matlID <= MaxAllowableID)) Then
                  ret = ThisMaterial.Get(ThisProperty.matlID)
                  MaterialTypeFilterPassed = CheckFilter(ThisMaterial.type,UseMaterialTypeFilter,MaterialTypeFilter)
                Else
                  MaterialTypeFilterPassed = (Not(UseMaterialTypeFilter) Or IncludeNoMaterialInMaterialTypeFilter)
                End If
                If (MaterialTypeFilterPassed) Then
                                'If the element passes both the type and material filters then add the related elements to a list of weld-related elements that must be reflected along with the weld
                  ret = ElemsOnWeldsToReflect.AddRule(ThisElem.connectSEG(0),FGD_ELEM_BYPROP)
                  ret = ElemsOnWeldsToReflect.AddRule(ThisElem.connectSEG(1),FGD_ELEM_BYPROP)
                Else
                                'If the element does not pass the material filter then warn the user that this weld element cannot be reflected because it uses a material type that the user has opted not to reflect
                  WarningMsgCount = WarningMsgCount+1
                  App.feAppMessage(FCM_WARNING,"WARNING: Weld element " & ThisElem.ID & " cannot be reflected because it is defined on a property that does not pass the current material type filter.")
                  ret = WeldElementsToRemove.Add(ThisElem.ID)
                End If
              Else
                                'If the element does not pass the element type filter then warn the user that this weld element cannot be reflected because it uses a property that the user has opted not to reflect
                  WarningMsgCount = WarningMsgCount+1
                  App.feAppMessage(FCM_WARNING,"WARNING: Weld element " & ThisElem.ID & " cannot be reflected because it is defined on a property that does not pass the current element type filter.")
                  ret = WeldElementsToRemove.Add(ThisElem.ID)
              End If
          End Select
        End If
        ret = ElemsToReflect.Next
      Wend
                                'Remove all non-reflectable weld elements from the set of elements that will be reflected
      ret = ElemsToReflect.RemoveSet(WeldElementsToRemove.ID)
                                'Check if there are any elements that must be reflected along with selected weld elements that are not already selected for reflection
      ret = ElemsOnWeldsToReflect.RemoveSet(ElemsToReflect.ID)
      If (ElemsOnWeldsToReflect.Count > 0) Then
                                'If so then tell the user that those elements must be reflected along with the weld
        ret = App.feAppMessageBox(2,"Weld elements have been selected for reflection that reference other unselected elements. Add referenced elements?" & vbCrLf & vbCrLf & "Yes - welds will be reflected along with their referenced elements" & vbCrLf & "No - these welds will not be reflected")
        If (ret = FE_OK) Then
                                'If the user accepts then add the referenced elements to the set of elements to reflect
          ret = ElemsToReflect.AddSet(ElemsOnWeldsToReflect.ID)
        Else
                                'If the user does not accept then remove from the reflection set the weld elements which reference other elements that the user does not want to reflect
          WeldElementsToRemove.Clear
          ret = ElemsToReflect.First
          While(ret <> FE_FAIL)
            ret = ThisElem.Get(ElemsToReflect.CurrentID)
            If (ThisElem.topology = FTO_WELD2) Then
            Select Case ThisElem.connectTYPE(0)
              Case 0,3,7 'Elem to Elem, Patch to Patch, Fastener to Elem
              	If ((ElemsToReflect.IsAdded(ThisElem.connectSEG(0)) = FE_FAIL) Or (ElemsToReflect.IsAdded(ThisElem.connectSEG(1)) = FE_FAIL)) Then
                  ret = WeldElementsToRemove.Add(ThisElem.ID)
                End If
              Case 1 'Elem to Elem Vertex
              	If (ElemsToReflect.IsAdded(ThisElem.connectSEG(0)) = FE_FAIL) Then
                  ret = WeldElementsToRemove.Add(ThisElem.ID)
                End If
            End Select
            End If
            ret = ElemsToReflect.Next
          Wend
          ret = ElemsToReflect.RemoveSet(WeldElementsToRemove.ID)
        End If
      End If
                                'Use the set of elements to reflect to create a set of nodes that must be reflected along with the elements
      NodesToReflect.Clear
      ret = ElemsToReflect.First
      While(ret <> FE_FAIL)
      	ret = ThisElem.Get(ElemsToReflect.CurrentID)
                                'Get all the nodes that are in any way associated with each element that is to be reflected
        Call GetNodesToReflectForElem(App,ThisElem,NodesToReflectForThisElem)
        ret = NodesToReflectForThisElem.First
                                'Add each associated node to the set of nodes to reflect if it is not already added
        While (ret <> FE_FAIL)
          ret = NodesToReflect.IsAdded(NodesToReflectForThisElem.CurrentID)
          If (ret = FE_FAIL) Then
            ret = NodesToReflect.Add(NodesToReflectForThisElem.CurrentID)
          End If
          ret = NodesToReflectForThisElem.Next
        Wend
        ret = ElemsToReflect.Next
      Wend
                                'Use the set of elements to reflect and the filter selections to create a set of properties that will be reflected along with the elements
      PropertiesToReflect.Clear
      ret = ElemsToReflect.First
      While (ret <> FE_FAIL)
        ret = ThisElem.Get(ElemsToReflect.CurrentID)
                                'Check each element type to see if it passes the type filter for reflected property creation
        If (CheckFilter(ThisElem.type,UseElemTypeFilter,ElemTypeFilter)) Then
                                'Do not try to add the property if it does not exist
          If ((ThisElem.propID > 0) And (ThisElem.propID <= MaxAllowableID)) Then
                                'Do not bother checking further if the property is already added to the set
            If (PropertiesToReflect.IsAdded(ThisElem.propID) = FE_FAIL) Then
              ret = ThisProperty.Get(ThisElem.propID)
                                'If the element type passes the type filter then check the material type to see if it passes the material type filter for reflected property creation
              If ((ThisProperty.matlID > 0) And (ThisProperty.matlID <= MaxAllowableID)) Then
                ret = ThisMaterial.Get(ThisProperty.matlID)
                MaterialTypeFilterPassed = CheckFilter(ThisMaterial.type,UseMaterialTypeFilter,MaterialTypeFilter)
              Else
                MaterialTypeFilterPassed = (Not(UseMaterialTypeFilter) Or IncludeNoMaterialInMaterialTypeFilter)
              End If
              If (MaterialTypeFilterPassed) Then
                                'If the element passes both the element type filter and the material type filter than add its property to the set of properties that will be reflected
                ret = PropertiesToReflect.Add(ThisElem.propID)
              End If
            End If
          End If
                                'If the element is a weld that references one or more properties then add those properties to the set of properties to reflect as long as the material type filter is passed
          If (ThisElem.topology = FTO_WELD2) Then
            If ((ThisElem.connectTYPE(0) = 4) Or (ThisElem.connectTYPE(0) = 8)) Then
              If (PropertiesToReflect.IsAdded(ThisElem.connectSEG(0)) = FE_FAIL) Then
                ret = ThisProperty.Get(ThisElem.connectSEG(0))
                If ((ThisProperty.matlID > 0) And (ThisProperty.matlID <= MaxAllowableID)) Then
                  ret = ThisMaterial.Get(ThisProperty.matlID)
                  MaterialTypeFilterPassed = CheckFilter(ThisMaterial.type,UseMaterialTypeFilter,MaterialTypeFilter)
                Else
                  MaterialTypeFilterPassed = (Not(UseMaterialTypeFilter) Or IncludeNoMaterialInMaterialTypeFilter)
                End If
                If (MaterialTypeFilterPassed) Then
                  ret = PropertiesToReflect.Add(ThisElem.connectSEG(0))
                End If
              End If
              If (PropertiesToReflect.IsAdded(ThisElem.connectSEG(1)) = FE_FAIL) Then
                ret = ThisProperty.Get(ThisElem.connectSEG(1))
                If ((ThisProperty.matlID > 0) And (ThisProperty.matlID <= MaxAllowableID)) Then
                  ret = ThisMaterial.Get(ThisProperty.matlID)
                  MaterialTypeFilterPassed = CheckFilter(ThisMaterial.type,UseMaterialTypeFilter,MaterialTypeFilter)
                Else
                  MaterialTypeFilterPassed = (Not(UseMaterialTypeFilter) Or IncludeNoMaterialInMaterialTypeFilter)
                End If
                If (MaterialTypeFilterPassed) Then
                  ret = PropertiesToReflect.Add(ThisElem.connectSEG(1))
                End If
              End If
            End If
          End If
        End If
        ret = ElemsToReflect.Next
      Wend
                                'The ID of zero cannot be added to a set, so use a separate boolean variable to indicate whether CS0 should be reflected
      ReflectCS0 = False
                                'Use the set of properties to reflect to create a set of coordinate systems that will be reflected along with the properties
      CSysToReflect.Clear
      ret = PropertiesToReflect.First
      While (ret <> FE_FAIL)
        ret = ThisProperty.Get(PropertiesToReflect.CurrentID)
                                'Check if this property to be reflected will need CS0 to be reflected
        If (ThisProperty.refCS = 0) Then
          ReflectCS0 = True
                                'Check if this property to be reflected will need any other coordinate system to be reflected and if so then add it to the set of coordinate systems to reflect if it is not already added
        ElseIf ((ThisProperty.refCS > 0) And (ThisProperty.refCS <= MaxAllowableID)) Then
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: Adding CSys " & ThisProperty.refCS & " to reflection list.")
          End If
          If (CSysToReflect.IsAdded(ThisProperty.refCS) = FE_FAIL) Then
            ret = CSysToReflect.Add(ThisProperty.refCS)
          End If
        End If
        ret = PropertiesToReflect.Next
      Wend
                                'If the user wishes to create reflected nodal coordinate systems then use the set of nodes to reflect to add to the set of coordinate systems that will be reflected along with the properties
      If (ReflectNodeCSys) Then
        ret = NodesToReflect.First
        While (ret <> FE_FAIL)
          ret = ThisNode.Get(NodesToReflect.CurrentID)
            If (ThisNode.defCSys = 0) Then
              ReflectCS0 = True
            ElseIf (CSysToReflect.IsAdded(ThisNode.defCSys) = FE_FAIL) Then
              ret = CSysToReflect.Add(ThisNode.defCSys)
            End If
            If (ThisNode.outCSys = 0) Then
              ReflectCS0 = True
            ElseIf (CSysToReflect.IsAdded(ThisNode.outCSys) = FE_FAIL) Then
              ret = CSysToReflect.Add(ThisNode.outCSys)
            End If
          ret = NodesToReflect.Next
        Wend
      End If
                                'Reflect all of the coordinate systems in the set of coordinate systems to reflect and create a lookup list relating each of the reflected coordinate systems to the coordinate system from which it was reflected
      ReflectionCSysLookup.Clear
                                'CS0 is a special case because it has ID zero and it is "basic" and cannot be retrieved like user created coordinate systems
      If (ReflectCS0) Then
                                'Get the ID for the reflection of CS0
        If (NewCSysOffsetID < 3) Then
          NewCSysID = ThisCsys.NextEmptyID
        Else
          NewCSysID = NewCSysOffsetID
          ret = TestCsysID.Get(NewCSysID)
          If (ret <> FE_FAIL) Then
            NewCSysID = ThisCsys.NextEmptyID
            WarningMsgCount = WarningMsgCount+1
            App.feAppMessage(FCM_WARNING,"WARNING: Offset CSys ID " & ThisCsys.ID & " is already in use; the next available ID " & NewCSysID & " will be used instead.")
          End If
        End If
        If (DebugMode) Then
          App.feAppMessage(FCM_NORMAL,"DEBUG: Reflecting CSys 0 to CSys " & NewCSysID & ".")
        End If
                                'Create a reflection of CS0
        ret = ThisCsys.Put(NewCSysID)
        ret = ThisCsys.Get(NewCSysID)
                                'CS0 is "basic" so its origin is always (0.0, 0.0, 0.0)
        PointToReflect(0) = 0.0
        PointToReflect(1) = 0.0
        PointToReflect(2) = 0.0
                                'Reflect the origin
        Call ReflectPoint(PointToReflect,SymPlaneNormal,SymPlanePoint,ReflectedPoint)
                                'CS0 is "basic" so its X-axis is always (1.0, 0.0, 0.0)
        XaxisToReflect(0) = 1.0
        XaxisToReflect(1) = 0.0
        XaxisToReflect(2) = 0.0
                                'CS0 is "basic" so its Y-axis is always (0.0, 1.0, 0.0)
        YaxisToReflect(0) = 0.0
        YaxisToReflect(1) = 1.0
        YaxisToReflect(2) = 0.0
        If (DebugMode) Then
          App.feAppMessage(FCM_NORMAL,"DEBUG: X-axis before reflection is (" & XaxisToReflect(0) & ", " & XaxisToReflect(1) & ", " & XaxisToReflect(2) & ")")
          App.feAppMessage(FCM_NORMAL,"DEBUG: Y-axis before reflection is (" & YaxisToReflect(0) & ", " & YaxisToReflect(1) & ", " & YaxisToReflect(2) & ")")
        End If
                                'Reflect the X-axis
        Call ReflectVector(XaxisToReflect,SymPlaneNormal,ReflectedXaxis)
                                'Reflect the Y-axis
        Call ReflectVector(YaxisToReflect,SymPlaneNormal,ReflectedYaxis)
                                'Compute the Z-axis as the cross product of the reflected X-axis with the reflected Y-axis
        Call VectorCrossProduct(ReflectedXaxis,ReflectedYaxis,ReflectedZaxis)
        If (DebugMode) Then
          App.feAppMessage(FCM_NORMAL,"DEBUG: X-axis after reflection is (" & ReflectedXaxis(0) & ", " & ReflectedXaxis(1) & ", " & ReflectedXaxis(2) & ")")
          App.feAppMessage(FCM_NORMAL,"DEBUG: Y-axis after reflection is (" & ReflectedYaxis(0) & ", " & ReflectedYaxis(1) & ", " & ReflectedYaxis(2) & ")")
          App.feAppMessage(FCM_NORMAL,"DEBUG: Z-axis after reflection is (" & ReflectedZaxis(0) & ", " & ReflectedZaxis(1) & ", " & ReflectedZaxis(2) & ")")
        End If
                                'Compute the rotation angles for the reflected coordinate system
        ret = ComputeXYZangles(App,ReflectedXaxis,ReflectedYaxis,ReflectedZaxis,CsysXYZrotations)
                                'Assign the origin to the reflected coordinate system
        ThisCsys.origin = ReflectedPoint
                                'Assign the rotation angles to the reflected coordinate system
        ThisCsys.rotx = CsysXYZrotations(0)
        ThisCsys.roty = CsysXYZrotations(1)
        ThisCsys.rotz = CsysXYZrotations(2)
                                'Assign the type of the reflected coordinate system to be the same as CS0
        ThisCsys.type = FCS_RECTANGULAR
                                'The ID zero cannot be added to the reflected coordinate system lookup list so store the reflected CS0 ID in a separate variable
        ReflectedCS0ID = NewCSysID
                                'Update the database with the new reflected coodinate system
        ret = ThisCsys.Put(NewCSysID)
      End If
                                'All coordinate systems to reflect other than CS0 are stored in the set of coordinate systems to reflect
      ret = CSysToReflect.First
      While (ret <> FE_FAIL)
                                'CS1 and CS2 are special cases because they are "basic" and cannot be retrieved like user created coordinate systems
        If ((CSysToReflect.CurrentID = 1) Or (CSysToReflect.CurrentID = 2)) Then
                                'Get the ID for the reflection of CS1 or CS2
          If (NewCSysOffsetID < 3) Then
            NewCSysID = ThisCsys.NextEmptyID
          Else
            NewCSysID = CSysToReflect.CurrentID+NewCSysOffsetID
            ret = TestCsysID.Get(NewCSysID)
            If (ret <> FE_FAIL) Then
              NewCSysID = ThisCsys.NextEmptyID
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Offset CSys ID " & ThisCsys.ID & " is already in use; the next available ID " & NewCSysID & " will be used instead.")
            End If
          End If
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: Reflecting CSys " & CSysToReflect.CurrentID & " to CSys " & NewCSysID & ".")
          End If
                                'Create a reflection of CS1 or CS2
          ret = ThisCsys.Put(NewCSysID)
          ret = ThisCsys.Get(NewCSysID)
                                'CS1 and CS2 are "basic" so their origin is always (0.0, 0.0, 0.0)
          PointToReflect(0) = 0.0
          PointToReflect(1) = 0.0
          PointToReflect(2) = 0.0
                                'Reflect the origin
          Call ReflectPoint(PointToReflect,SymPlaneNormal,SymPlanePoint,ReflectedPoint)
                                'CS1 and CS2 are "basic" so their X-axis is always (1.0, 0.0, 0.0)
          XaxisToReflect(0) = 1.0
          XaxisToReflect(1) = 0.0
          XaxisToReflect(2) = 0.0
                                'CS1 and CS2 are "basic" so their Y-axis is always (0.0, 1.0, 0.0)
          YaxisToReflect(0) = 0.0
          YaxisToReflect(1) = 1.0
          YaxisToReflect(2) = 0.0
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: X-axis before reflection is (" & XaxisToReflect(0) & ", " & XaxisToReflect(1) & ", " & XaxisToReflect(2) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Y-axis before reflection is (" & YaxisToReflect(0) & ", " & YaxisToReflect(1) & ", " & YaxisToReflect(2) & ")")
          End If
                                'Reflect the X-axis
          Call ReflectVector(XaxisToReflect,SymPlaneNormal,ReflectedXaxis)
                                'Reflect the Y-axis
          Call ReflectVector(YaxisToReflect,SymPlaneNormal,ReflectedYaxis)
                                'Compute the Z-axis as the cross product of the reflected X-axis with the reflected Y-axis
          Call VectorCrossProduct(ReflectedXaxis,ReflectedYaxis,ReflectedZaxis)
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: X-axis after reflection is (" & ReflectedXaxis(0) & ", " & ReflectedXaxis(1) & ", " & ReflectedXaxis(2) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Y-axis after reflection is (" & ReflectedYaxis(0) & ", " & ReflectedYaxis(1) & ", " & ReflectedYaxis(2) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Z-axis after reflection is (" & ReflectedZaxis(0) & ", " & ReflectedZaxis(1) & ", " & ReflectedZaxis(2) & ")")
          End If
                                'Compute the rotation angles for the reflected coordinate system
          ret = ComputeXYZangles(App,ReflectedXaxis,ReflectedYaxis,ReflectedZaxis,CsysXYZrotations)
                                'Assign the origin to the reflected coordinate system
          ThisCsys.origin = ReflectedPoint
                                'Assign the rotation angles to the reflected coordinate system
          ThisCsys.rotx = CsysXYZrotations(0)
          ThisCsys.roty = CsysXYZrotations(1)
          ThisCsys.rotz = CsysXYZrotations(2)
                                'Assign the type of the reflected coordinate system to be the same as CS1(cylindrical) or CS2(spherical)
          If (CSysToReflect.CurrentID = 1) Then
            ThisCsys.type = FCS_CYLINDRICAL
          ElseIf (CSysToReflect.CurrentID = 2) Then
            ThisCsys.type = FCS_SPHERICAL
          End If
                                'Add an entry to the reflected coordinate system lookup list for the reflection of CS1 or CS2
          ret = ReflectionCSysLookup.AddInt(CSysToReflect.CurrentID,NewCSysID,0,0)
                                'Update the database with the new reflected coodinate system
          ret = ThisCsys.Put(NewCSysID)
          If (DebugMode) Then
            ret = ThisCsys.Get(NewCSysID)
            App.feAppMessage(FCM_NORMAL,"DEBUG: Csys X-axis after reflection is (" & ThisCsys.matrix(0) & ", " & ThisCsys.matrix(1) & ", " & ThisCsys.matrix(2) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Csys Y-axis after reflection is (" & ThisCsys.matrix(3) & ", " & ThisCsys.matrix(4) & ", " & ThisCsys.matrix(5) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Csys Z-axis after reflection is (" & ThisCsys.matrix(6) & ", " & ThisCsys.matrix(7) & ", " & ThisCsys.matrix(8) & ")")
          End If
        Else
                                'If the ID >= 3 then the coordinate system is a user coordinate system not a "basic" coordinate system
          ret = ThisCsys.Get(CSysToReflect.CurrentID)
                                'Get the ID for the reflection of the current user coordinate system
          If (NewCSysOffsetID < 3) Then
            NewCSysID = ThisCsys.NextEmptyID
          Else
            NewCSysID = ThisCsys.ID+NewCSysOffsetID
            ret = TestCsysID.Get(NewCSysID)
            If (ret <> FE_FAIL) Then
              NewCSysID = ThisCsys.NextEmptyID
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Offset CSys ID " & ThisCsys.ID & " is already in use; the next available ID " & NewCSysID & " will be used instead.")
            End If
          End If
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: Reflecting CSys " & CSysToReflect.CurrentID & " to CSys " & NewCSysID & ".")
          End If
                                'Get the origin of the user coordinate system
          PointToReflect(0) = ThisCsys.origin(0)
          PointToReflect(1) = ThisCsys.origin(1)
          PointToReflect(2) = ThisCsys.origin(2)
                                'Reflect the origin
          Call ReflectPoint(PointToReflect,SymPlaneNormal,SymPlanePoint,ReflectedPoint)
                                'Get the X-axis of the user coordinate system
          XaxisToReflect(0) = ThisCsys.matrix(0)
          XaxisToReflect(1) = ThisCsys.matrix(1)
          XaxisToReflect(2) = ThisCsys.matrix(2)
                                'Get the Y-axis of the user coordinate system
          YaxisToReflect(0) = ThisCsys.matrix(3)
          YaxisToReflect(1) = ThisCsys.matrix(4)
          YaxisToReflect(2) = ThisCsys.matrix(5)
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: X-axis before reflection is (" & XaxisToReflect(0) & ", " & XaxisToReflect(1) & ", " & XaxisToReflect(2) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Y-axis before reflection is (" & YaxisToReflect(0) & ", " & YaxisToReflect(1) & ", " & YaxisToReflect(2) & ")")
          End If
                                'Reflect the X-axis
          Call ReflectVector(XaxisToReflect,SymPlaneNormal,ReflectedXaxis)
                                'Reflect the Y-axis
          Call ReflectVector(YaxisToReflect,SymPlaneNormal,ReflectedYaxis)
                                'Compute the Z-axis as the cross product of the reflected X-axis with the reflected Y-axis
          Call VectorCrossProduct(ReflectedXaxis,ReflectedYaxis,ReflectedZaxis)
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: X-axis after reflection is (" & ReflectedXaxis(0) & ", " & ReflectedXaxis(1) & ", " & ReflectedXaxis(2) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Y-axis after reflection is (" & ReflectedYaxis(0) & ", " & ReflectedYaxis(1) & ", " & ReflectedYaxis(2) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Z-axis after reflection is (" & ReflectedZaxis(0) & ", " & ReflectedZaxis(1) & ", " & ReflectedZaxis(2) & ")")
          End If
                                'Compute the rotation angles for the reflected coordinate system
          ret = ComputeXYZangles(App,ReflectedXaxis,ReflectedYaxis,ReflectedZaxis,CsysXYZrotations)
                                'Assign the origin to the reflected coordinate system
          ThisCsys.origin = ReflectedPoint
                                'Assign the rotation angles to the reflected coordinate system
          ThisCsys.rotx = CsysXYZrotations(0)
          ThisCsys.roty = CsysXYZrotations(1)
          ThisCsys.rotz = CsysXYZrotations(2)
                                'Add an entry to the reflected coordinate system lookup list for the reflection of the current user coordinate system
          ret = ReflectionCSysLookup.AddInt(CSysToReflect.CurrentID,NewCSysID,0,0)
                                'Update the database with the new reflected coodinate system
          ret = ThisCsys.Put(NewCSysID)
          If (DebugMode) Then
            ret = ThisCsys.Get(NewCSysID)
            App.feAppMessage(FCM_NORMAL,"DEBUG: Csys X-axis after reflection is (" & ThisCsys.matrix(0) & ", " & ThisCsys.matrix(1) & ", " & ThisCsys.matrix(2) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Csys Y-axis after reflection is (" & ThisCsys.matrix(3) & ", " & ThisCsys.matrix(4) & ", " & ThisCsys.matrix(5) & ")")
            App.feAppMessage(FCM_NORMAL,"DEBUG: Csys Z-axis after reflection is (" & ThisCsys.matrix(6) & ", " & ThisCsys.matrix(7) & ", " & ThisCsys.matrix(8) & ")")
          End If
        End If
                                'Get the next coordinate system to reflect
        ret = CSysToReflect.Next
      Wend
                                'Reflect all of the nodes in the set of nodes to reflect and create a lookup list relating each of the reflected nodes to the node from which it was reflected
      ReflectionNodeLookup.Clear
      ret = NodesToReflect.First
      While (ret <> FE_FAIL)
        ret = ThisNode.Get(NodesToReflect.CurrentID)
                                'Get the ID for the reflection of the current node
        If (NewNodeOffsetID = 0) Then
          NewNodeID = ThisNode.NextEmptyID
        Else
          NewNodeID = ThisNode.ID+NewNodeOffsetID
          ret = TestNodeID.Get(NewNodeID)
          If (ret <> FE_FAIL) Then
            NewNodeID = ThisNode.NextEmptyID
            WarningMsgCount = WarningMsgCount+1
            App.feAppMessage(FCM_WARNING,"WARNING: Offset node ID " & ThisNode.ID+NewNodeOffsetID & " is already in use; the available ID " & NewNodeID & " will be used instead.")
          End If
        End If
                                'Get the coordinates of the node to reflect
        PointToReflect(0) = ThisNode.x
        PointToReflect(1) = ThisNode.y
        PointToReflect(2) = ThisNode.z
                                'Reflect the coordinates
        Call ReflectPoint(PointToReflect,SymPlaneNormal,SymPlanePoint,ReflectedPoint)
                                'Assign the coordinates of the reflected node
        ThisNode.x = ReflectedPoint(0)
        ThisNode.y = ReflectedPoint(1)
        ThisNode.z = ReflectedPoint(2)
                                'If the user opted to reflect nodal coordinate systems then assign reflected coordinate systems using the reflected coordinate system lookup list
        If (ReflectNodeCSys) Then
          If (ThisNode.defCSys = 0) Then
            ThisNode.defCSys = ReflectedCS0ID
          Else
            ret = ReflectionCSysLookup.IsAdded(ThisNode.defCSys)
            If (ret = FE_FAIL) Then
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Reflected CSys lookup failure on node " & ThisNode.ID & " for CSys " & ThisNode.defCSys & ". This unreflected CSys will remain assigned to the node.")
            Else
              ret = ReflectionCSysLookup.Find(ThisNode.defCSys)
              ThisNode.defCSys = ReflectionCSysLookup.nInt1
            End If
          End If
          If (ThisNode.outCSys = 0) Then
            ThisNode.outCSys = ReflectedCS0ID
          Else
            ret = ReflectionCSysLookup.IsAdded(ThisNode.outCSys)
            If (ret = FE_FAIL) Then
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Reflected CSys lookup failure on node " & ThisNode.ID & " for CSys " & ThisNode.outCSys & ". This unreflected CSys will remain assigned to the node.")
            Else
              ret = ReflectionCSysLookup.Find(ThisNode.outCSys)
              ThisNode.outCSys = ReflectionCSysLookup.nInt1
            End If
          End If
        End If
                                'Add an entry to the reflected node lookup list for the reflection of the current node
        ret = ReflectionNodeLookup.AddInt(NodesToReflect.CurrentID,NewNodeID,0,0)
                                'Update the database with the new reflected node
        ret = ThisNode.Put(NewNodeID)
                                'Get the next node to reflect
        ret = NodesToReflect.Next
      Wend
                                'Reflect all of the properties in the set of properties to reflect and create a lookup list relating each of the reflected properties to the property from which it was reflected
      ReflectionPropertyLookup.Clear
      ret = PropertiesToReflect.First
      While (ret <> FE_FAIL)
        ret = ThisProperty.Get(PropertiesToReflect.CurrentID)
                                'Make sure the property references a valid coordinate system - otherwise it cannot be reflected
        If ((ThisProperty.refCS < 0) Or (ThisProperty.refCS > MaxAllowableID)) Then
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: Skipping property " & PropertiesToReflect.CurrentID & " because its refCSys is " & ThisProperty.refCS & ".")
          End If
          ret = ReflectionPropertyLookup.AddInt(PropertiesToReflect.CurrentID,PropertiesToReflect.CurrentID,0,0)
        Else
                                'Get the ID for the reflection of the current property
          If (NewPropOffsetID = 0) Then
            NewPropID = ThisProperty.NextEmptyID
          Else
            NewPropID = ThisProperty.ID+NewPropOffsetID
            ret = TestPropertyID.Get(NewPropID)
            If (ret <> FE_FAIL) Then
              NewPropID = ThisProperty.NextEmptyID
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Offset property ID " & ThisProperty.ID+NewPropOffsetID & " is already in use; the available ID " & NewPropID & " will be used instead.")
            End If
          End If
                                'Assign the appropriate reflected coordinate system to the reflected property
          If (ThisProperty.refCS = 0) Then
                                'Reflected CS0 is stored in a separate variable
            ThisProperty.refCS = ReflectedCS0ID
          Else
                                'All other reflected coordinate systems are in the reflected coordinate system lookup list
            ret = ReflectionCSysLookup.IsAdded(ThisProperty.refCS)
            If (ret = FE_FAIL) Then
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Reflected CSys lookup failure on property " & ThisProperty.ID & " for CSys " & ThisProperty.refCS & ". This property will not be reflected.")
            Else
              ret = ReflectionCSysLookup.Find(ThisProperty.refCS)
              ThisProperty.refCS = ReflectionCSysLookup.nInt1
            End If
          End If
                                'Add an entry to the reflected property list for the reflection of the current property
          ret = ReflectionPropertyLookup.AddInt(PropertiesToReflect.CurrentID,NewPropID,0,0)
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: Reflecting property " & PropertiesToReflect.CurrentID & " to property " & NewPropID & ".")
          End If
                                'Update the database with the new reflected property
          ret = ThisProperty.Put(NewPropID)
        End If
                                'Get the next node to reflect
        ret = PropertiesToReflect.Next
      Wend
                                'Reflect all of the elements in the set of elements to reflect
      ret = ElemsToReflect.First
      While (ret <> FE_FAIL)
        ret = ThisElem.Get(ElemsToReflect.CurrentID)
                                'Get the ID for the reflection of the current element
        If (NewElemOffsetID = 0) Then
          NewElemID = ThisElem.NextEmptyID
        Else
          NewElemID = ThisElem.ID+NewElemOffsetID
          ret = TestElemID.Get(NewElemID)
          If (ret <> FE_FAIL) Then
            NewElemID = ThisElem.NextEmptyID
            WarningMsgCount = WarningMsgCount+1
            App.feAppMessage(FCM_WARNING,"WARNING: Offset element ID " & ThisElem.ID+NewElemOffsetID & " is already in use; the available ID " & NewElemID & " will be used instead.")
          End If
        End If
                                'If there is a problem with reflecting the current element then set a flag variable
        ReflectionError = False
                                'Reflect elements by replacing the original node assignments with reflected nodes from the reflected node lookup list
        Select Case ThisElem.topology
          Case FTO_POINT1,FTO_LINE2,FTO_TETRA4,FTO_WEDGE6,FTO_BRICK8,FTO_TETRA10,FTO_WEDGE15,FTO_BRICK20
                                'For most element topologies only the nodes array needs to be updated
            GetNodeIndecesForElemTopology(ThisElem.topology,NodeIndeces)
            ret = NodeIndeces.First
            While (ret <> FE_FAIL)
              ret = ReflectionNodeLookup.IsAdded(ThisElem.Node(NodeIndeces.CurrentID-1))
              If (ret = FE_FAIL) Then
                WarningMsgCount = WarningMsgCount+1
                App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & ThisElem.Node(NodeIndeces.CurrentID-1) & ". This element will not be reflected.")
                ReflectionError = True
              Else
                ret = ReflectionNodeLookup.Find(ThisElem.Node(NodeIndeces.CurrentID-1))
                ThisElem.Node(NodeIndeces.CurrentID-1) = ReflectionNodeLookup.nInt1
              End If
              ret = NodeIndeces.Next
            Wend
          Case FTO_TRIA3,FTO_TRIA6,FTO_QUAD4,FTO_QUAD8
                                'For thin shell types, the reflected node order must be reversed to preserve the correct front/back side
            OriginalNodeArray = Empty
            OriginalNodeArray = ThisElem.Nodes
            GetNodeIndecesForElemTopology(ThisElem.topology,NodeIndeces)
            NodeIndecesArray = Empty
            ret = NodeIndeces.GetArray(DummySize,NodeIndecesArray)
            For ArrayPosition = LBound(NodeIndecesArray) To UBound(NodeIndecesArray)
              ret = ReflectionNodeLookup.IsAdded(OriginalNodeArray(NodeIndecesArray(ArrayPosition)-1))
              If (ret = FE_FAIL) Then
                WarningMsgCount = WarningMsgCount+1
                App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & OriginalNodeArray(NodeIndecesArray(ArrayPosition)-1) & ". This element will not be reflected.")
                ReflectionError = True
              Else
                ret = ReflectionNodeLookup.Find(OriginalNodeArray(NodeIndecesArray(ArrayPosition)-1))
                ThisElem.Node(NodeIndecesArray(UBound(NodeIndecesArray)+LBound(NodeIndecesArray)-ArrayPosition)-1) = ReflectionNodeLookup.nInt1
              End If
            Next ArrayPosition
          Case FTO_RIGIDLIST
                                'For rigid elements, nodes are in the node array and in one node list
            GetNodeIndecesForElemTopology(ThisElem.topology,NodeIndeces)
            ret = NodeIndeces.First
            While (ret <> FE_FAIL)
              ret = ReflectionNodeLookup.IsAdded(ThisElem.Node(NodeIndeces.CurrentID-1))
              If (ret = FE_FAIL) Then
                WarningMsgCount = WarningMsgCount+1
                App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & ThisElem.Node(NodeIndeces.CurrentID-1) & ". This element will not be reflected.")
                ReflectionError = True
              Else
                ret = ReflectionNodeLookup.Find(ThisElem.Node(NodeIndeces.CurrentID-1))
                ThisElem.Node(NodeIndeces.CurrentID-1) = ReflectionNodeLookup.nInt1
              End If
              ret = NodeIndeces.Next
            Wend
            ret = ThisElem.GetNodeList(0,NumberOfNodesInNodeList,NodeListNodeArray,NodeListFaceArray,NodeListWeightArray,NodeListDOFarray)
            If ((ret = FE_NOT_EXIST) Or (ret = FE_NOT_AVAILABLE)) Then
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Node list not found on element " & ThisElem.ID & ". This element will not be reflected.")
              ReflectionError = True
            Else
              For ArrayPosition = LBound(NodeListNodeArray) To UBound(NodeListNodeArray)
                ret = ReflectionNodeLookup.IsAdded(NodeListNodeArray(ArrayPosition))
                If (ret = FE_FAIL) Then
                  WarningMsgCount = WarningMsgCount+1
                  App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & NodeListNodeArray(ArrayPosition) & ". This element will not be reflected.")
                  ReflectionError = True
                Else
                  ret = ReflectionNodeLookup.Find(NodeListNodeArray(ArrayPosition))
                  NodeListNodeArray(ArrayPosition) = ReflectionNodeLookup.nInt1
                End If
              Next ArrayPosition
              If (Not ReflectionError) Then
                ret = ThisElem.PutNodeList(0,NumberOfNodesInNodeList,NodeListNodeArray,NodeListFaceArray,NodeListWeightArray,NodeListDOFarray)
              End If
            End If
          Case FTO_MULTILIST2
                                'For multilist elements, nodes are in two node lists
            ret = ThisElem.GetNodeList(0,NumberOfNodesInNodeList,NodeListNodeArray,NodeListFaceArray,NodeListWeightArray,NodeListDOFarray)
            If ((ret = FE_NOT_EXIST) Or (ret = FE_NOT_AVAILABLE)) Then
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Node list not found on element " & ThisElem.ID & ". This element will not be reflected.")
              ReflectionError = True
            Else
              For ArrayPosition = LBound(NodeListNodeArray) To UBound(NodeListNodeArray)
                ret = ReflectionNodeLookup.IsAdded(NodeListNodeArray(ArrayPosition))
                If (ret = FE_FAIL) Then
                  WarningMsgCount = WarningMsgCount+1
                  App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & NodeListNodeArray(ArrayPosition) & ". This element will not be reflected.")
                  ReflectionError = True
                Else
                  ret = ReflectionNodeLookup.Find(NodeListNodeArray(ArrayPosition))
                  NodeListNodeArray(ArrayPosition) = ReflectionNodeLookup.nInt1
                End If
              Next ArrayPosition
              If (Not ReflectionError) Then
                ret = ThisElem.PutNodeList(0,NumberOfNodesInNodeList,NodeListNodeArray,NodeListFaceArray,NodeListWeightArray,NodeListDOFarray)
              End If
            End If
            ret = ThisElem.GetNodeList(1,NumberOfNodesInNodeList,NodeListNodeArray,NodeListFaceArray,NodeListWeightArray,NodeListDOFarray)
            If ((ret = FE_NOT_EXIST) Or (ret = FE_NOT_AVAILABLE)) Then
              WarningMsgCount = WarningMsgCount+1
              App.feAppMessage(FCM_WARNING,"WARNING: Node list not found on element " & ThisElem.ID & ". This element will not be reflected.")
              ReflectionError = True
            Else
              For ArrayPosition = LBound(NodeListNodeArray) To UBound(NodeListNodeArray)
                ret = ReflectionNodeLookup.IsAdded(NodeListNodeArray(ArrayPosition))
                If (ret = FE_FAIL) Then
                  WarningMsgCount = WarningMsgCount+1
                  App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & NodeListNodeArray(ArrayPosition) & ". This element will not be reflected.")
                  ReflectionError = True
                Else
                  ret = ReflectionNodeLookup.Find(NodeListNodeArray(ArrayPosition))
                  NodeListNodeArray(ArrayPosition) = ReflectionNodeLookup.nInt1
                End If
              Next ArrayPosition
              If (Not ReflectionError) Then
                ret = ThisElem.PutNodeList(1,NumberOfNodesInNodeList,NodeListNodeArray,NodeListFaceArray,NodeListWeightArray,NodeListDOFarray)
              End If
            End If
          Case FTO_CONTACT2
                                'Contact elements do not seem to exist in this context
            WarningMsgCount = WarningMsgCount+1
            App.feAppMessage(FCM_WARNING,"WARNING: Contact2 not supported.")
            ReflectionError = True
          Case FTO_WELD2
                                'For weld elements, nodes are in the node array and possibly also in the connectSEG array
            For WeldElementNodeIndex = 0 To 19
              If (ThisElem.Node(WeldElementNodeIndex) > 0) Then
                ret = ReflectionNodeLookup.IsAdded(ThisElem.Node(WeldElementNodeIndex))
                If (ret = FE_FAIL) Then
                  WarningMsgCount = WarningMsgCount+1
                  App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & ThisElem.Node(WeldElementNodeIndex) & ". This element will not be reflected.")
                  ReflectionError = True
                Else
                  ret = ReflectionNodeLookup.Find(ThisElem.Node(WeldElementNodeIndex))
                  ThisElem.Node(WeldElementNodeIndex) = ReflectionNodeLookup.nInt1
                End If
              End If
            Next WeldElementNodeIndex
                                'For weld elements of type Elem vertex to Elem vertex only, the connectSEG array contains nodes
            If (ThisElem.connectTYPE(0) = 2) Then
              ret = ReflectionNodeLookup.IsAdded(ThisElem.connectSEG(0))
              If (ret = FE_FAIL) Then
                WarningMsgCount = WarningMsgCount+1
                App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & ThisElem.connectSEG(0) & ". This element will not be reflected.")
                ReflectionError = True
              Else
                ret = ReflectionNodeLookup.Find(ThisElem.connectSEG(0))
                ThisElem.connectSEG(0) = ReflectionNodeLookup.nInt1
              End If
              ret = ReflectionNodeLookup.IsAdded(ThisElem.connectSEG(1))
              If (ret = FE_FAIL) Then
                WarningMsgCount = WarningMsgCount+1
                App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & ThisElem.connectSEG(1) & ". This element will not be reflected.")
                ReflectionError = True
              Else
                ret = ReflectionNodeLookup.Find(ThisElem.connectSEG(1))
                ThisElem.connectSEG(1) = ReflectionNodeLookup.nInt1
              End If
            End If
          Case Else
            WarningMsgCount = WarningMsgCount+1
            App.feAppMessage(FCM_WARNING,"WARNING: Unknown topology type for element " & ThisElem.ID & ". This element will not be reflected.")
            ReflectionError = True
        End Select
                                'For all element types, check for an orientation node and assign to it the correct reflected node using the reflected node lookup list
        If (ThisElem.orientID > 0) Then
          ret = ReflectionNodeLookup.IsAdded(ThisElem.orientID)
          If (ret = FE_FAIL) Then
            WarningMsgCount = WarningMsgCount+1
            App.feAppMessage(FCM_WARNING,"WARNING: Reflected node lookup failure on element " & ThisElem.ID & " for node " & ThisElem.orientID & ". This element will not be reflected.")
            ReflectionError = True
          Else
            ret = ReflectionNodeLookup.Find(ThisElem.orientID)
            ThisElem.orientID = ReflectionNodeLookup.nInt1
          End If
        End If
                                'Reflect the orientation vector for all elements except shell elements because shell elements store material orientation angle instead
        If ((ThisElem.topology <> FTO_TRIA3) And (ThisElem.topology <> FTO_TRIA6) And (ThisElem.topology <> FTO_QUAD4) And (ThisElem.topology <> FTO_QUAD8)) Then
          VectorToReflect(0) = ThisElem.orient(0)
          VectorToReflect(1) = ThisElem.orient(1)
          VectorToReflect(2) = ThisElem.orient(2)
          Call ReflectVector(VectorToReflect,SymPlaneNormal,ReflectedVector)
          ThisElem.orient(0) = ReflectedVector(0)
          ThisElem.orient(1) = ReflectedVector(1)
          ThisElem.orient(2) = ReflectedVector(2)
        End If
                                'Reflect both offset vectors for all element types
        VectorToReflect(0) = ThisElem.offset(0,0)
        VectorToReflect(1) = ThisElem.offset(0,1)
        VectorToReflect(2) = ThisElem.offset(0,2)
        Call ReflectVector(VectorToReflect,SymPlaneNormal,ReflectedVector)
        ThisElem.offset(0,0) = ReflectedVector(0)
        ThisElem.offset(0,1) = ReflectedVector(1)
        ThisElem.offset(0,2) = ReflectedVector(2)
        VectorToReflect(0) = ThisElem.offset(1,0)
        VectorToReflect(1) = ThisElem.offset(1,1)
        VectorToReflect(2) = ThisElem.offset(1,2)
        Call ReflectVector(VectorToReflect,SymPlaneNormal,ReflectedVector)
        ThisElem.offset(1,0) = ReflectedVector(0)
        ThisElem.offset(1,1) = ReflectedVector(1)
        ThisElem.offset(1,2) = ReflectedVector(2)
                                'Check each element type to see if it passes the type filter for reflected properties
        If (CheckFilter(ThisElem.type,UseElemTypeFilter,ElemTypeFilter)) Then
                                'Check each element material type to see if it passes the type filter for reflected properties
          If (ThisElem.propID > 0) Then
            If (PropertiesToReflect.IsAdded(ThisElem.propID) = FE_OK) Then
              ret = ThisProperty.Get(ThisElem.propID)
              If (ThisProperty.matlID > 0) Then
                ret = ThisMaterial.Get(ThisProperty.matlID)
                MaterialTypeFilterPassed = CheckFilter(ThisMaterial.type,UseMaterialTypeFilter,MaterialTypeFilter)
              Else
                MaterialTypeFilterPassed = (Not(UseMaterialTypeFilter) Or IncludeNoMaterialInMaterialTypeFilter)
              End If
              If (MaterialTypeFilterPassed) Then
                                'If the element passes both the type and material filters then replace the property assignment with its reflected property from the reflected property lookup list
                ret = ReflectionPropertyLookup.IsAdded(ThisElem.propID)
                If (ret = FE_FAIL) Then
                  WarningMsgCount = WarningMsgCount+1
                  App.feAppMessage(FCM_WARNING,"WARNING: Reflected property lookup failure on element " & ThisElem.ID & " for property " & ThisElem.propID & ". This element will not be reflected.")
                  ReflectionError = True
                Else
                  ret = ReflectionPropertyLookup.Find(ThisElem.propID)
                  If (DebugMode) Then
                    App.feAppMessage(FCM_NORMAL,"DEBUG: Assigning reflected property " & ReflectionPropertyLookup.nInt1 & " to reflected element " & NewElemID & ".")
                  End If
                  ThisElem.propID = ReflectionPropertyLookup.nInt1
                End If
              End If
            End If
          End If
                                'For weld elements only, check to see if additional properties are referenced and need to be replaced by reflected properties
          If (ThisElem.topology = FTO_WELD2) Then
                                'For weld elements of type Prop To Prop or Fastener To Prop only, replace the properties in the connectSEG array with reflected ones as long as the material type filter is passed
            If ((ThisElem.connectTYPE(0) = 4) Or (ThisElem.connectTYPE(0) = 8)) Then
              If (PropertiesToReflect.IsAdded(ThisElem.connectSEG(0)) = FE_OK) Then
                ret = ThisProperty.Get(ThisElem.connectSEG(0))
                If ((ThisProperty.matlID > 0) And (ThisProperty.matlID <= MaxAllowableID)) Then
                  ret = ThisMaterial.Get(ThisProperty.matlID)
                  MaterialTypeFilterPassed = CheckFilter(ThisMaterial.type,UseMaterialTypeFilter,MaterialTypeFilter)
                Else
                  MaterialTypeFilterPassed = (Not(UseMaterialTypeFilter) Or IncludeNoMaterialInMaterialTypeFilter)
                End If
                If (MaterialTypeFilterPassed) Then
                  ret = ReflectionPropertyLookup.IsAdded(ThisElem.connectSEG(0))
                  If (ret = FE_FAIL) Then
                    WarningMsgCount = WarningMsgCount+1
                    App.feAppMessage(FCM_WARNING,"WARNING: Reflected property lookup failure on element " & ThisElem.ID & " for property " & ThisElem.connectSEG(0) & ". This element will not be reflected.")
                    ReflectionError = True
                  Else
                    ret = ReflectionPropertyLookup.Find(ThisElem.connectSEG(0))
                    If (DebugMode) Then
                      App.feAppMessage(FCM_NORMAL,"DEBUG: Assigning reflected property " & ReflectionPropertyLookup.nInt1 & " to reflected weld property " & NewElemID & ".")
                    End If
                    ThisElem.connectSEG(0) = ReflectionPropertyLookup.nInt1
                  End If
                End If
              End If
              If (PropertiesToReflect.IsAdded(ThisElem.connectSEG(1)) = FE_OK) Then
                ret = ThisProperty.Get(ThisElem.connectSEG(1))
                If ((ThisProperty.matlID > 0) And (ThisProperty.matlID <= MaxAllowableID)) Then
                  ret = ThisMaterial.Get(ThisProperty.matlID)
                  MaterialTypeFilterPassed = CheckFilter(ThisMaterial.type,UseMaterialTypeFilter,MaterialTypeFilter)
                Else
                  MaterialTypeFilterPassed = (Not(UseMaterialTypeFilter) Or IncludeNoMaterialInMaterialTypeFilter)
                End If
                If (MaterialTypeFilterPassed) Then
                  ret = ReflectionPropertyLookup.IsAdded(ThisElem.connectSEG(1))
                  If (ret = FE_FAIL) Then
                    WarningMsgCount = WarningMsgCount+1
                    App.feAppMessage(FCM_WARNING,"WARNING: Reflected property lookup failure on element " & ThisElem.ID & " for property " & ThisElem.connectSEG(1) & ". This element will not be reflected.")
                    ReflectionError = True
                  Else
                    ret = ReflectionPropertyLookup.Find(ThisElem.connectSEG(1))
                    If (DebugMode) Then
                      App.feAppMessage(FCM_NORMAL,"DEBUG: Assigning reflected property " & ReflectionPropertyLookup.nInt1 & " to reflected weld property " & NewElemID & ".")
                    End If
                    ThisElem.connectSEG(1) = ReflectionPropertyLookup.nInt1
                  End If
                End If
              End If
            End If
          End If
        End If
                                'If no errors occured with the element reflection then update the database with the new reflected element
        If (Not(ReflectionError)) Then
          If (DebugMode) Then
            App.feAppMessage(FCM_NORMAL,"DEBUG: Reflecting element " & ThisElem.ID & " to element " & NewElemID & ".")
          End If
          ret = ThisElem.Put(NewElemID)
        End If
                                'Get the next element to reflect
        ret = ElemsToReflect.Next
      Wend
                                'Tell the user that processing is complete after a load definition has been created for the last group in the sequence
      App.feAppMessage(FCM_NORMAL," ")
      App.feAppMessage(FCM_NORMAL,"Model processing complete.")
                                'Tell the user if any warning messages were issued at runtime
      If (WarningMsgCount = 0) Then
        App.feAppMessage(FCM_NORMAL,"No warning messages were issued.")
      ElseIf (WarningMsgCount = 1) Then
        App.feAppMessage(FCM_WARNING,"1 warning message was issued - Please see message output for details.")
      Else
        App.feAppMessage(FCM_WARNING,WarningMsgCount & " warning messages were issued - Please see message output for details.")
      End If
    End If
  End If
End If
If (ret = FE_CANCEL) Then
                                'If the user cancels at any point, give a message to that effect and terminate
  App.feAppMessage(FCM_NORMAL,"User cancelled. No modifications were made to the model.")
End If
End Sub

Function CheckFilter(ByVal CheckID As Long, ByVal HasFilter As Boolean, ByRef FilterSet As femap.Set) As Boolean
  If (HasFilter) Then
                                'If a filter is active then check the ID to see if it is incuded in the filter
    CheckFilter = (FilterSet.IsAdded(CheckID) = FE_OK)
  Else
                                'If a filter is not active then anything passes
    CheckFilter = True
  End If
End Function

Sub GetNodesToReflectForElem(ByRef ThisApp As femap.model, ByRef ElemToReflect As femap.Elem, ByRef NodesToReflectForElem As femap.Set)
                                'Return a set with all of the nodes associated with a particular element
  Dim NodeIndecesForElem As femap.Set
  Set NodeIndecesForElem = ThisApp.feSet
  Dim NodeListForElem As femap.Set
  Set NodeListForElem = ThisApp.feSet
  Dim WeldNodeIndex As Long
  NodesToReflectForElem.Clear
  If (ElemToReflect.topology = FTO_WELD2) Then
                                'Weld type elements are a special case
    For WeldNodeIndex = 0 To 19
                                'Get any nodes stored in the node array of a weld element
      If (ElemToReflect.Node(WeldNodeIndex) > 0) Then
        ReturnCode = NodesToReflectForElem.IsAdded(ElemToReflect.Node(WeldNodeIndex))
        If (ReturnCode = FE_FAIL) Then
          ReturnCode = NodesToReflectForElem.Add(ElemToReflect.Node(WeldNodeIndex))
        End If
      End If
    Next WeldNodeIndex
                                'For weld elements of type Elem vertex to Elem vertex only, add the two vertex nodes
    If (ElemToReflect.connectTYPE(0) = 2) Then
      ReturnCode = NodesToReflectForElem.IsAdded(ElemToReflect.connectSEG(0))
      If (ReturnCode = FE_FAIL) Then
        ReturnCode = NodesToReflectForElem.Add(ElemToReflect.connectSEG(0))
      End If
      ReturnCode = NodesToReflectForElem.IsAdded(ElemToReflect.connectSEG(1))
      If (ReturnCode = FE_FAIL) Then
        ReturnCode = NodesToReflectForElem.Add(ElemToReflect.connectSEG(1))
      End If
    End If
  Else
                                'For non-weld elements get the nodes from the node array according to topology
    Call GetNodeIndecesForElemTopology(ElemToReflect.topology,NodeIndecesForElem)
    ReturnCode = NodeIndecesForElem.First
    While (ReturnCode <> FE_FAIL)
      ReturnCode = NodesToReflectForElem.Add(ElemToReflect.Node(NodeIndecesForElem.CurrentID-1))
      ReturnCode = NodeIndecesForElem.Next
    Wend
                                'For non-weld elements add any nodes stored in node lists
    Call GetNodeListForElem(ElemToReflect,NodeListForElem)
    ReturnCode = NodesToReflectForElem.AddSet(NodeListForElem.ID)
  End If
                                'For any element that uses an orientation node add it as well
  If (ElemToReflect.orientID > 0) Then
    ReturnCode = NodesToReflectForElem.IsAdded(ElemToReflect.orientID)
    If (ReturnCode = FE_FAIL) Then
      ReturnCode = NodesToReflectForElem.Add(ElemToReflect.orientID)
    End If
  End If
End Sub

Sub GetNodeIndecesForElemTopology(ByVal elTopology As Long, ByRef NodeIndexSet As femap.Set)
                                'Return an array of node indeces for the node array of an element topology
  NodeIndexSet.Clear
  Dim DivByZero As Long
  Select Case elTopology
    Case FTO_POINT1 'Point
      ReturnCode = NodeIndexSet.AddArray(1,Array(1))
    Case FTO_LINE2 'Line
      ReturnCode = NodeIndexSet.AddArray(2,Array(1,2))
    Case FTO_TRIA3 'Tri3
      ReturnCode = NodeIndexSet.AddArray(3,Array(1,2,3))
    Case FTO_TRIA6 'Tri6
      ReturnCode = NodeIndexSet.AddArray(6,Array(1,2,3,5,6,7))
    Case FTO_QUAD4 'Quad4
      ReturnCode = NodeIndexSet.AddArray(4,Array(1,2,3,4))
    Case FTO_QUAD8 'Quad8
      ReturnCode = NodeIndexSet.AddArray(8,Array(1,2,3,4,5,6,7,8))
    Case FTO_TETRA4 'Tetra4
      ReturnCode = NodeIndexSet.AddArray(4,Array(1,2,3,5))
    Case FTO_WEDGE6 'Wedge6
      ReturnCode = NodeIndexSet.AddArray(6,Array(1,2,3,5,6,7))
    Case FTO_BRICK8 'Brick8
      ReturnCode = NodeIndexSet.AddArray(8,Array(1,2,3,4,5,6,7,8))
    Case FTO_TETRA10 'Tetra10
      ReturnCode = NodeIndexSet.AddArray(10,Array(1,2,3,5,9,10,11,13,14,15))
    Case FTO_WEDGE15 'Wedge15
      ReturnCode = NodeIndexSet.AddArray(15,Array(1,2,3,5,6,7,9,10,11,13,14,15,17,18,19))
    Case FTO_BRICK20 'Brick20
      ReturnCode = NodeIndexSet.AddArray(20,Array(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20))
    Case FTO_RIGIDLIST 'Rigid
      ReturnCode = NodeIndexSet.AddArray(1,Array(1))
    Case FTO_MULTILIST2 'SlideLine
      ReturnCode = FE_OK
    Case FTO_WELD2 'Weld
                                'Weld elements are handled separately and should never be called here
      DivByZero = 0
                                'If a weld type element is called here then force an exception
      DivByZero = 1 \ DivByZero
  End Select
End Sub

Sub GetNodeListForElem(ByRef e As femap.Elem, ByRef NodeListSet As femap.Set)
                                'Return a set of nodes that are stored in node lists on an element
  Dim NumberOfNodes As Long
  Dim ArrayOfNodes As Variant
  Dim dummy_FaceArray As Variant
  Dim dummy_WeightArray As Variant
  Dim dummy_DOFarray As Variant
  Select Case e.topology
    Case FTO_POINT1 'Point
      NodeListSet.Clear
    Case FTO_LINE2 'Line
      NodeListSet.Clear
    Case FTO_TRIA3 'Tri3
      NodeListSet.Clear
    Case FTO_TRIA6 'Tri6
      NodeListSet.Clear
    Case FTO_QUAD4 'Quad4
      NodeListSet.Clear
    Case FTO_QUAD8 'Quad8
      NodeListSet.Clear
    Case FTO_TETRA4 'Tetra4
      NodeListSet.Clear
    Case FTO_WEDGE6 'Wedge6
      NodeListSet.Clear
    Case FTO_BRICK8 'Brick8
      NodeListSet.Clear
    Case FTO_TETRA10 'Tetra10
      NodeListSet.Clear
    Case FTO_WEDGE15 'Wedge15
      NodeListSet.Clear
    Case FTO_BRICK20 'Brick20
      NodeListSet.Clear
    Case FTO_RIGIDLIST 'Rigid
      NodeListSet.Clear
      ReturnCode = e.GetNodeList(0,NumberOfNodes,ArrayOfNodes,dummy_FaceArray,dummy_WeightArray,dummy_DOFarray)
      ReturnCode = NodeListSet.AddArray(NumberOfNodes,ArrayOfNodes)
    Case FTO_MULTILIST2 'SlideLine
      NodeListSet.Clear
      ReturnCode = e.GetNodeList(0,NumberOfNodes,ArrayOfNodes,dummy_FaceArray,dummy_WeightArray,dummy_DOFarray)
      ReturnCode = NodeListSet.AddArray(NumberOfNodes,ArrayOfNodes)
      ReturnCode = e.GetNodeList(1,NumberOfNodes,ArrayOfNodes,dummy_FaceArray,dummy_WeightArray,dummy_DOFarray)
      ReturnCode = NodeListSet.AddArray(NumberOfNodes,ArrayOfNodes)
    Case FTO_WELD2 'Weld
      NodeListSet.Clear
  End Select
End Sub

Sub ReflectVector(ByVal VectorComponents() As Double, ByVal PlaneNormal() As Double, ByRef ReflectedVector() As Double)
                                'Reflect a vector across a plane defined by a plane normal vector
  Dim UnitNormal(0 To 2) As Double
  Dim VectorMagnitude As Double
  Dim PerpendicularComponent As Double
  Dim PerpendicularVector(0 To 2) As Double
  Dim ParllelVector(0 To 2) As Double
  Dim CoordinateIndex As Long
  VectorMagnitude = Sqr(DotProduct(PlaneNormal,PlaneNormal))
  For CoordinateIndex = 0 To 2
    UnitNormal(CoordinateIndex) = PlaneNormal(CoordinateIndex)/VectorMagnitude
  Next CoordinateIndex
  PerpendicularComponent = DotProduct(VectorComponents,UnitNormal)
  For CoordinateIndex = 0 To 2
    PerpendicularVector(CoordinateIndex) = PerpendicularComponent*UnitNormal(CoordinateIndex)
    ParllelVector(CoordinateIndex) = VectorComponents(CoordinateIndex)-PerpendicularVector(CoordinateIndex)
    ReflectedVector(CoordinateIndex) = ParllelVector(CoordinateIndex)-PerpendicularVector(CoordinateIndex)
  Next CoordinateIndex
End Sub

Sub ReflectPoint(ByVal PointCoordinates() As Double, ByVal PlaneNormal() As Double, ByVal PlanePoint() As Double, ByRef ReflectedCoordinates() As Double)
                                'Reflect a point across a plane defined by a plane normal vector and a point that the plane passes through
  Dim UnitNormal(0 To 2) As Double
  Dim VectorMagnitude As Double
  Dim PlaneToPoint(0 To 2) As Double
  Dim DistanceFromPlane As Double
  Dim PerpendicularVector(0 To 2) As Double
  Dim CoordinateIndex As Long
  VectorMagnitude = Sqr(DotProduct(PlaneNormal,PlaneNormal))
  For CoordinateIndex = 0 To 2
    UnitNormal(CoordinateIndex) = PlaneNormal(CoordinateIndex)/VectorMagnitude
    PlaneToPoint(CoordinateIndex) = PointCoordinates(CoordinateIndex)-PlanePoint(CoordinateIndex)
  Next CoordinateIndex
  DistanceFromPlane = DotProduct(PlaneToPoint,UnitNormal)
  For CoordinateIndex = 0 To 2
    PerpendicularVector(CoordinateIndex) = DistanceFromPlane*UnitNormal(CoordinateIndex)
    ReflectedCoordinates(CoordinateIndex) = PointCoordinates(CoordinateIndex)-2*PerpendicularVector(CoordinateIndex)
  Next CoordinateIndex
End Sub

Function DotProduct(ByVal VectorA() As Double, ByVal VectorB() As Double) As Double
                                'Compute the dot product of two vectors
  DotProduct = VectorA(0)*VectorB(0)+VectorA(1)*VectorB(1)+VectorA(2)*VectorB(2)
End Function

Sub VectorCrossProduct(ByVal Vec1() As Double,ByVal Vec2() As Double, Vec1CrossVec2() As Double)
                                'Compute the cross product of two vectors
  Vec1CrossVec2(0) = Vec1(1)*Vec2(2)-Vec1(2)*Vec2(1)
  Vec1CrossVec2(1) = Vec1(2)*Vec2(0)-Vec1(0)*Vec2(2)
  Vec1CrossVec2(2) = Vec1(0)*Vec2(1)-Vec1(1)*Vec2(0)
End Sub

Function ComputeXYZangles(ByRef ThisApp As femap.model, ByVal xvec() As Double, ByVal yvec() As Double, ByVal zvec() As Double, rotationXYZangles() As Double) As zReturnCode
                                'Compute the X followed by Y followed by Z rotation angles for a local coordinate system rotated from global
  Dim ReturnCode As zReturnCode
  ReturnCode = FE_OK
  If (Abs(zvec(0)) > unity+unitlength_tolerance) Then
    ReturnCode = FE_FAIL
    ThisApp.feAppMessage(FCM_ERROR,"ERROR: Internal data check failure")
  ElseIf (Abs(zvec(0)) > unity-unitlength_tolerance) Then
    rotationXYZangles(0) = 0.0
    rotationXYZangles(1) = ArcsinDegrees(zvec(0))
    rotationXYZangles(2) = ArctanDegrees(xvec(1),yvec(1))
  Else
    rotationXYZangles(0) = ArctanDegrees(-zvec(1),zvec(2))
    rotationXYZangles(1) = ArcsinDegrees(zvec(0))
    rotationXYZangles(2) = ArctanDegrees(-yvec(0),xvec(0))
  End If
  ComputeXYZangles = ReturnCode
End Function

Function ArctanDegrees(ByVal num As Double, ByVal den As Double) As Double
                                'Return the arctangent of a number in degrees
  ArctanDegrees = (45.0/Atn(unity))*myArctan(num,den)
End Function

Function ArcsinDegrees(ByVal x As Double) As Double
                                'Return the arcsine of a number in degrees
  ArcsinDegrees = (45.0/Atn(unity))*myArcsin(x)
End Function

Function myArctan(ByVal num As Double, ByVal den As Double) As Double
                                'Compute the four-quadrant arctangent of a number represented by the fraction num/den
  If (den > 0.0) Then
  	myArctan = Atn(num/den)
  ElseIf (den = 0.0) Then
    If (num < 0.0) Then
      myArctan = -2.0*Atn(unity)
    Else
      myArctan = 2.0*Atn(unity)
    End If
  Else
    If (num < 0.0) Then
      myArctan = -4.0*Atn(unity)+Atn(num/den)
    Else
      myArctan = 4.0*Atn(unity)+Atn(num/den)
    End If
  End If
End Function

Function myArcsin(ByVal x As Double) As Double
                                'Compute the arcsine of a number
  If (Abs(Abs(x)-unity) < unitlength_tolerance) Then
    myArcsin = myArctan(x,0.0)
  Else
    myArcsin = myArctan(x,Sqr(unity-x*x))
  End If
End Function
